from flaschengeist.database import db
from flaschengeist.models import ModelSerializeMixin

from typing import Optional

drink_tag_association = db.Table(
    "drink_x_tag",
    db.Column("drink_id", db.Integer, db.ForeignKey("drink.id")),
    db.Column("tag_id", db.Integer, db.ForeignKey("drink_tag.id")),
)

drink_type_association = db.Table(
    "drink_x_type",
    db.Column("drink_id", db.Integer, db.ForeignKey("drink.id")),
    db.Column("type_id", db.Integer, db.ForeignKey("drink_type.id")),
)


class Tag(db.Model, ModelSerializeMixin):
    """
    Tag
    """

    __tablename__ = "drink_tag"
    id: int = db.Column("id", db.Integer, primary_key=True)
    name: str = db.Column(db.String(30), nullable=False, unique=True)


class DrinkType(db.Model, ModelSerializeMixin):
    """
    DrinkType
    """

    __tablename__ = "drink_type"
    id: int = db.Column("id", db.Integer, primary_key=True)
    name: str = db.Column(db.String(30), nullable=False, unique=True)


class DrinkPrice(db.Model, ModelSerializeMixin):
    """
    PriceFromVolume
    """

    __tablename__ = "drink_price"
    id: int = db.Column("id", db.Integer, primary_key=True)
    volume: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False), nullable=False)
    price: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False))
    drink_id = db.Column("drink_id", db.Integer, db.ForeignKey("drink.id"))
    drink = db.relationship("Drink", back_populates="prices")
    no_auto: bool = db.Column(db.Boolean, default=False)
    public: bool = db.Column(db.Boolean, default=True)
    description: Optional[str] = db.Column(db.String(30))
    round_step: float = db.Column(db.Numeric(precision=3, scale=2, asdecimal=False), nullable=False, default=0.5)


class Ingredient(db.Model, ModelSerializeMixin):
    """
    Drink Build
    """

    __tablename__ = "drink_ingredient"
    id: int = db.Column("id", db.Integer, primary_key=True)
    volume: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False), nullable=False)
    drink_parent_id: int = db.Column("drink_parent_id", db.Integer, db.ForeignKey("drink.id"))
    drink_parent = db.relationship("Drink", foreign_keys=drink_parent_id)
    drink_ingredient_id: int = db.Column("drink_ingredient_id", db.Integer, db.ForeignKey("drink.id"))
    drink_ingredient: "Drink" = db.relationship("Drink", foreign_keys=drink_ingredient_id)


class Drink(db.Model, ModelSerializeMixin):
    """
    DrinkPrice
    """

    __tablename__ = "drink"
    id: int = db.Column("id", db.Integer, primary_key=True)
    name: str = db.Column(db.String(60), nullable=False)
    volume: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False))
    cost_price: float = db.Column(db.Numeric(precision=5, scale=2, asdecimal=False))
    discount: float = db.Column(db.Numeric(precision=3, scale=2, asdecimal=False), nullable=False)
    extra_charge: Optional[float] = db.Column(db.Numeric(precision=3, scale=2, asdecimal=False), default=0)
    prices: [DrinkPrice] = db.relationship(
        "DrinkPrice", back_populates="drink", cascade="all,delete,delete-orphan", order_by=[DrinkPrice.volume]
    )
    ingredients: [Ingredient] = db.relationship(
        "Ingredient", back_populates="drink_parent", foreign_keys=Ingredient.drink_parent_id
    )
    tags: [Optional[Tag]] = db.relationship("Tag", secondary=drink_tag_association, cascade="save-update, merge")
    type_id_ = db.Column("type_id", db.Integer, db.ForeignKey("drink_type.id"))
    type = db.relationship("DrinkType")
