from werkzeug.exceptions import BadRequest, NotFound
from sqlalchemy.exc import IntegrityError

from flaschengeist import logger
from flaschengeist.database import db
from .models import Drink, DrinkPrice, Ingredient, Tag, DrinkType

from math import ceil


def update():
    db.session.commit()


def get_tags():
    return Tag.query.all()


def get_tag(identifier):
    if isinstance(identifier, int):
        retVal = Tag.query.get(identifier)
    elif isinstance(identifier, str):
        retVal = Tag.query.filter(Tag.name == identifier).one_or_none()
    else:
        logger.debug("Invalid identifier type for Tag")
        raise BadRequest
    if not retVal:
        raise NotFound
    return retVal


def create_tag(name):
    try:
        tag = Tag(name=name)
        db.session.add(tag)
        update()
        return tag
    except IntegrityError:
        raise BadRequest("Name already exists")


def rename_tag(identifier, new_name):
    tag = get_tag(identifier)
    tag.name = new_name
    try:
        update()
    except IntegrityError:
        raise BadRequest("Name already exists")


def delete_tag(identifier):
    tag = get_tag(identifier)
    db.session.delete(tag)
    try:
        update()
    except IntegrityError:
        raise BadRequest("Tag still in use")


def get_drink_types():
    return DrinkType.query.all()


def get_drink_type(identifier):
    if isinstance(identifier, int):
        retVal = DrinkType.query.get(identifier)
    elif isinstance(identifier, str):
        retVal = DrinkType.query.filter(Tag.name == identifier).one_or_none()
    else:
        logger.debug("Invalid identifier type for DrinkType")
        raise BadRequest
    if not retVal:
        raise NotFound
    return retVal


def create_drink_type(name):
    try:
        drinkType = DrinkType(name=name)
        db.session.add(drinkType)
        update()
        return drinkType
    except IntegrityError:
        raise BadRequest("Name already exists")


def rename_drink_type(identifier, new_name):
    drink_type = get_drink_type(identifier)
    drink_type.name = new_name
    try:
        update()
    except IntegrityError:
        raise BadRequest("Name already exists")
    return drink_type


def delete_drink_type(identifier):
    drinkType = get_drink_type(identifier)
    db.session.delete(drinkType)
    try:
        update()
    except IntegrityError:
        raise BadRequest("DrinkType still in use")


def round_price(price, round_step):
    return round(ceil(float(price) / round_step) * round_step * 100) / 100


def calc_prices(drink, prices):
    retVal = []
    if len(drink.ingredients) > 0:
        return calc_price_by_ingredients(drink, prices)
    allowed_keys = DrinkPrice().serialize().keys()
    for price in prices:
        values = {key: value for key, value in price.items() if key in allowed_keys}
        if values.get("no_auto"):
            retVal.append(DrinkPrice(**values))
        else:
            volume = float(values.pop("volume"))
            if "price" in values:
                values.pop("price")
            _price = float(drink.cost_price) / float(drink.volume) * volume
            _price += _price * float(drink.discount)
            if drink.extra_charge:
                _price += float(drink.extra_charge)
            _price = round_price(_price, float(price.get("round_step")))
            retVal.append(DrinkPrice(volume=volume, price=_price, **values))
    return retVal


def calc_price_by_ingredients(drink, prices):
    allowed_keys = DrinkPrice().serialize().keys()
    retVal = []
    for price in prices:
        values = {key: value for key, value in price.items() if key in allowed_keys}
        if values.get("no_auto"):
            retVal.append(DrinkPrice(**values))
        else:
            volume = float(values.pop("volume"))
            if "price" in values:
                values.pop("price")
            _price = 0
            for ingredient in drink.ingredients:
                _price = (
                    float(ingredient.drink_ingredient.cost_price)
                    / float(ingredient.drink_ingredient.volume)
                    * float(ingredient.volume)
                )
            _price += _price * float(drink.discount) + float(drink.extra_charge)
            _price = round_price(_price, price.get("round_step"))
            retVal.append(DrinkPrice(volume=volume, price=_price, **values))
    return retVal


def get_drinks(name=None):
    if name:
        return Drink.query.filter(Drink.name.contains(name)).all()
    return Drink.query.all()


def get_drink(identifier):
    if isinstance(identifier, int):
        retVal = Drink.query.get(identifier)
    elif isinstance(identifier, str):
        retVal = Drink.query.filter(Tag.name == identifier).one_or_none()
    else:
        logger.debug("Invalid identifier type for Drink")
        raise BadRequest
    if not retVal:
        raise NotFound
    return retVal


def add_prices(drink, prices):
    for price in prices:
        drink.prices.append(price)


def add_ingredients(drink, ingredients):
    for identifier, volume in ingredients:
        ingredient = Ingredient(volume=volume, drink_ingredient=get_drink(identifier))
        drink.ingredients.append(ingredient)


def create_drink(data):
    allowed_keys = Drink().serialize().keys()
    values = {key: value for key, value in data.items() if key in allowed_keys}
    prices = values.pop("prices", [])
    ingredients = values.pop("ingredients", [])
    if "id" in values:
        values.pop("id")

    drink = Drink(**values)
    add_ingredients(drink, ingredients)
    drink.prices = calc_prices(drink, prices)
    db.session.add(drink)
    update()
    return drink


def delete_drink(identifier):
    drink = get_drink(identifier)
    for price in drink.prices:
        db.session.delete(price)
    for ingredient in drink.ingredients:
        db.session.delete(ingredient)
    db.session.delete(drink)
    update()
